import numpy as np
import matplotlib.pyplot as plt

from typing import List, Dict, Union, Optional

# NOTE - Based on: https://towardsdatascience.com/451f07999373
# NOTE - Does not work, use only as a reference

def generate_architecture(input_layer_size, hidden_layer_info,
    output_layer_size, activations=dict()):
    input_layer = dict()
    input_layer["layer_size"] = input_layer_size
    input_layer["activation"] = activations.get("input", "none")

    hidden_layers = list()
    hidden_size = len(hidden_layer_info)
    hidden_layer_activations = activations.get("hidden", ["relu"] * hidden_size)
    for h, hidden_layer_size in enumerate(hidden_layer_info):
        hidden_layer = dict()
        hidden_layer["layer_size"] = hidden_layer_size
        hidden_layer["activation"] = hidden_layer_activations[h]
        hidden_layers.append(hidden_layer)

    output_layer = dict()
    output_layer["layer_size"] = output_layer_size
    output_layer["activation"] = activations.get("output", "sigmoid")

    return [input_layer] + hidden_layers + [output_layer]

def initialize_parameters(arch, seed=3):
    np.random.seed(seed)
    parameters = dict()
    layer_count = len(arch)

    for l in range(1, layer_count):
        layer_size = arch[l]["layer_size"]
        prev_layer_size = arch[l-1]["layer_size"]
        init_weight = np.random.randn(layer_size, prev_layer_size) * 0.01
        init_bias = np.zeros((arch[l]["layer_size"], 1))
        parameters["W" + str(l)] = init_weight
        parameters["b" + str(l)] = init_bias

    return parameters

def sigmoid(z):
    return 1 / (1 + np.exp(-z))

def sigmoid_d(da, z):
    s = sigmoid(z)
    ds = s * (1 - s)
    return da * ds

def relu(z):
    return np.maximum(0, z)

def relu_d(da, z):
    dz = np.array(da, copy=True)
    dz[Z <= 0] = 0
    return dz

def linear_forward(a, w, b):
    return np.dot(w, a) + b

def linear_activation_forward(prev_a, w, b, activation):
    z = linear_forward(prev_a, w, b)
    if activation == "sigmoid":
        a = sigmoid(z)

    elif activation == "relu":
        a = relu(z)

    return z, a

def model_forward(x, parameters, arch):
    forward_cache = dict()
    a = x
    layer_count = len(arch)

    for l in range(1, layer_count):
        prev_a = a
        w = parameters["W" + str(l)]
        b = parameters["b" + str(l)]
        activation = arch[l]["activation"]
        z, a = linear_activation_forward(prev_a, w, b, activation)
        forward_cache["Z" + str(l)] = Z
        forward_cache["A" + str(l-1)] = a

    al = a

    return al, forward_cache 

def compute_cost(al, y):
    m = y.shape[1]
    logs = np.multiply(np.log(al), y) + np.multiply(1 - y, np.log(1 - al))
    cost = np.squeeze(-np.sum(logs) / m)

    return cost

def linear_backward(dz, prev_a, w):
    m = prev_a.shape[1]
    dw = np.dot(dz, prev_a.T) / m
    db = np.sum(dz, axis=1, keepdims=True) / m
    prev_da = np.dot(w.T, dz)

    return prev_da, dw, db

def linear_activation_backward(da, z, prev_a, w, activation):
    if activation == "sigmoid":
        dz = sigmoid_d(da, z)

    elif activation == "relu":
        dz = relu_d(da, z)

    return linear_backward(dz, prev_a, w)

def model_backward(al, y, parameters, forward_cache, arch):
    grads = dict()
    layer_count = len(arch)
    m = al.shape[1]
    y = y.reshape(al.shape)

    dal = -(np.divide(y, al) - np.divide(1 - y, 1 - al))
    prev_da = dal

    for l in reversed(range(1, layer_count)):
        curr_da = prev_da
        activation = arch[l]["activation"]
        curr_w = parameters["W" + str(l)]
        curr_z = parameters["Z" + str(l)]
        prev_a = parameters["A" + str(l-1)]

        args = curr_da, curr_z, prev_a, curr_w, activation
        result = linear_activation_backward(*args)
        prev_da, curr_dw, curr_db = result

        grads["dW" + str(l)] = curr_dw
        grads["db" + str(l)] = curr_db

    return grads

def update_parameters(parameters, grads, learning_rate):
    layer_count = len(parameters) // 2
    for l in range(1, layer_count):
        offset_weight = learning_rate * grads["dW" + str(l)]
        new_weight = parameters["W" + str(l)] - offset_weight
        parameters["W" + str(l)] = new_weight

        offset_bias = learning_rate * grads["db" + str(l)]
        new_bias = parameters["b" + str(l)] - offset_bias
        parameters["b" + str(l)] = new_bias

    return parameters

def layer_model(x, y, arch, learning_rate=0.005, loops=3000, seed=1, show=False):
    np.random.seed(seed)

    costs = list()
    parameters = initialize_parameters(arch)

    for i in range(0, loops):
        al, forward_cache = model_forward(x, parameters, arch)
        cost = compute_cost(al, y)
        grads = model_backword(al, y, paramters, forward_cache, arch)
        parameters = update_parameters(parameters, grads, learning_rate)

        costs.append(cost)
        if show and i % 100 == 0:
            epoch = i / 100
            print(f"Epoch {epoch}: {cost}")

    plt.plot(np.squeeze(costs))
    plt.ylabel("cost")
    plt.xlabel("loops")
    plt.title(f"Learning Rate = {learning_rate}")
    plt.show()

    return parameters
